﻿using System;
using System.Configuration;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Linq;
using System.ServiceProcess;
using System.Text;
using System.IO.Ports;
using System.Data.Common;
using System.Data.SqlClient;
using System.Threading;
using System.Globalization;
using System.ServiceModel;
using System.ServiceModel.Description;
using DrHeet.SportsRadar.ServiceContracts;
using DrHeet.SR3600Manager;


namespace DrHeet.SportsRadarDCS
{
    public partial class DCS : ServiceBase
    {
        private SR3600Manager.SR3600Manager srMgr;

        private int errorCount = 0;
        private int dataPoints = 0;
        private float calibrationFactor;
        protected int ERRORLIMIT = 10;

        private ServiceHost wcfHost;        // WCF ServiceHost hosting the speed event subscription server

        private SubscriptionServer speedServer = new SubscriptionServer();

        private CultureInfo culture;        // for SR3600 number format parsing

        public DCS()
        {
            InitializeComponent();
            srMgr = new SR3600Manager.SR3600Manager();
        }


        /// <summary>Initializes the VK011Manager instance with configuration values.</summary>
        protected override void OnStart(string[] args)
        {
            Thread.Sleep(12000);  // allow debugger to connect
            dataPoints = 0;
            errorCount = 0;

            try
            {
                srMgr.PortName = ((NameValueCollection)ConfigurationManager.GetSection("DCSModules/SR3600"))["PortName"];
                srMgr.BaudRate = int.Parse(((NameValueCollection)ConfigurationManager.GetSection("DCSModules/SR3600"))["BaudRate"]);
                srMgr.DataBits = int.Parse(((NameValueCollection)ConfigurationManager.GetSection("DCSModules/SR3600"))["DataBits"]);

                switch (int.Parse(((NameValueCollection)ConfigurationManager.GetSection("DCSModules/SR3600"))["StopBits"]))
                {
                    case 0:
                        srMgr.StopBits = StopBits.None;
                        break;
                    case 2:
                        srMgr.StopBits = StopBits.Two;
                        break;
                    case 1:
                    default:
                        srMgr.StopBits = StopBits.One;
                        break;
                }

                switch (((NameValueCollection)ConfigurationManager.GetSection("DCSModules/SR3600"))["Parity"])
                {
                    case "Even":
                        srMgr.Parity = Parity.Even;
                        break;
                    case "Odd":
                        srMgr.Parity = Parity.Odd;
                        break;
                    case "None":
                    default:
                        srMgr.Parity = Parity.None;
                        break;
                }

                srMgr.CallBackHandler = SR3600Data;
                srMgr.OnError = SR3600Error;

                // make sure the locale used for number parsing is consistent between driver and service
                culture = new CultureInfo(ConfigurationManager.AppSettings.Get("Culture"));
                srMgr.Culture = culture;

                // The speed value is multiplied with this factor before it is saved
                calibrationFactor = float.Parse(((NameValueCollection)ConfigurationManager.GetSection("DCSModules/SR3600"))["CalibrationFactor"], culture);

                // Initialize data series
                OpenSR3600DataSeries();

                // Initialize serial communcation
                srMgr.InitSR3600();

                WriteToApplicationLog("SR3600 DCS ready. Listening on " + srMgr.PortName, EventLogEntryType.Information);
            }
            catch (Exception e)
            {
                WriteToApplicationLog("Error initializing application", EventLogEntryType.Error);
                WriteToApplicationLog(e.Message, EventLogEntryType.Error);
                this.Stop();
            }

            CreateServiceHost();
        }


        protected override void OnStop()
        {
            CloseSR3600DataSeries();
            srMgr.CleanupSR3600();
            RemoveServiceHost();
        }

        /// <summary>Updates channel information in the database from configuration values.</summary>
        private void OpenSR3600DataSeries()
        {
            string connStr = ConfigurationManager.ConnectionStrings["SportsRadarDCS"].ConnectionString;

            // we do not keep the connection open
            using (SqlConnection connection = new SqlConnection(connStr))
            {
                System.Data.SqlClient.SqlCommand cmd = new System.Data.SqlClient.SqlCommand();
                cmd.CommandType = System.Data.CommandType.Text;
                cmd.CommandText =
                    "INSERT INTO [DCS].[dbo].[SR3600DataSeries] " +
                    "  ([system_name]" +
                    "  ,[start_date]" +
                    "  ,[calibration_factor]" +
                    "  ,[channel_name]" +
                    "  ,[property_id]" +
                    "  ,[property_description])" +
                    " VALUES " +
                    " (@system_name " +
                    " ,@start_date " +
                    " ,@calibration_factor " +
                    " ,@channel_name " +
                    " ,@property_id " +
                    " ,@property_description) ";

                //paremeters are getting passed through the query

                cmd.Parameters.Add(new SqlParameter("@system_name", ConfigurationManager.AppSettings.Get("SystemName")));
                cmd.Parameters.Add(new SqlParameter("@start_date", DateTime.Now));
                cmd.Parameters.Add(new SqlParameter("@calibration_factor", calibrationFactor));
                cmd.Parameters.Add(new SqlParameter("@channel_name", ((NameValueCollection)ConfigurationManager.GetSection("DCSModules/SR3600"))["Channel"]));
                cmd.Parameters.Add(new SqlParameter("@property_id", ((NameValueCollection)ConfigurationManager.GetSection("DCSModules/SR3600"))["PropertyId"]));
                cmd.Parameters.Add(new SqlParameter("@property_description", ((NameValueCollection)ConfigurationManager.GetSection("DCSModules/SR3600"))["PropertyDescription"]));
                
                cmd.Connection = connection;

                try
                {
                    connection.Open();

                    if (cmd.ExecuteNonQuery() != 1)
                        SR3600Error("Error writing to the database.", null);
                }
                catch (Exception e)
                {
                    SR3600Error(e.Message, e);
                }
                finally
                {
                    connection.Close();
                }
            }
        }

        private void CreateServiceHost()
        {
            Uri baseAddress = new Uri( ((NameValueCollection)ConfigurationManager.GetSection("DCSModules/SR3600"))["WCFServiceHost"]);

            wcfHost = new ServiceHost(typeof(SubscriptionServer), baseAddress);

            try
            {
                // Add a service endpoint.
                wcfHost.AddServiceEndpoint(
                    typeof(ISubscriptionServer),
                    new WSDualHttpBinding(),
                    "SubscriptionServer");

                // Enable metadata exchange.
                ServiceMetadataBehavior smb = new ServiceMetadataBehavior();
                smb.HttpGetEnabled = true;
                wcfHost.Description.Behaviors.Add(smb);

                wcfHost.Open();
            }
            catch (CommunicationException ce)
            {
                WriteToApplicationLog(ce.Message, EventLogEntryType.Error);
                wcfHost.Abort();
            }
        }


        private void RemoveServiceHost()
        {
            try
            {
                wcfHost.Close();
            }
            catch (CommunicationException ce)
            {
                WriteToApplicationLog(ce.Message, EventLogEntryType.Error);
            }
        }


        private void CloseSR3600DataSeries()
        {
            string connStr = ConfigurationManager.ConnectionStrings["SportsRadarDCS"].ConnectionString;

            // we do not keep the connection open
            using (SqlConnection connection = new SqlConnection(connStr))
            {
                System.Data.SqlClient.SqlCommand cmd = new System.Data.SqlClient.SqlCommand();
                cmd.CommandType = System.Data.CommandType.Text;
                cmd.CommandText =
                    "UPDATE [DCS].[dbo].[SR3600DataSeries] " +
                    "  SET [end_date] = @end_date " +
                    "     ,[data_points] = @data_points " +
                    "WHERE [start_date] = (SELECT max([start_Date]) from [DCS].[dbo].[SR3600DataSeries])" +
                    "AND [end_date] is null ";

                //paremeters are getting passe through the query

                cmd.Parameters.Add(new SqlParameter("@end_date", DateTime.Now));
                cmd.Parameters.Add(new SqlParameter("@data_points", dataPoints));

                cmd.Connection = connection;

                try
                {
                    connection.Open();

                    if (cmd.ExecuteNonQuery() != 1)
                        SR3600Error("Error writing to the database.", null);
                }
                catch (Exception e)
                {
                    SR3600Error(e.Message, e);
                }
                finally
                {
                    connection.Close();
                }
            }
        }


        private void SR3600Data(SpeedRecord speed)
        {
            if (speed.speedValueMPH < 5.0 || speed.speedValueMPH > 100.0)
            {
                speed.speedValueMPH = -1.0;        // bad reading 
                speed.speedValueKPH = -1.0;        // bad reading 
            }
            else
            {
                speed.speedValueKPH = Math.Round(speed.speedValueKPH * calibrationFactor, 0);
                speed.speedValueMPH = Math.Round(speed.speedValueMPH * calibrationFactor, 0);
            }

            string connStr = ConfigurationManager.ConnectionStrings["SportsRadarDCS"].ConnectionString;

            // we do not keep the connection open
            using (SqlConnection connection = new SqlConnection(connStr))
            {
                System.Data.SqlClient.SqlCommand cmd = new System.Data.SqlClient.SqlCommand();
                cmd.CommandType = System.Data.CommandType.Text;
                cmd.CommandText =
                    "INSERT INTO [DCS].[dbo].[SR3600Data] " +
                    "  ([series_id]   " +
                    "  ,[dcs_time]   " +
                    "  ,[dcs_speed_mph]  " +
                    "  ,[dcs_speed_kph]) " +
                    " SELECT    " +
                    " (SELECT max([series_id]) from [DCS].[dbo].[SR3600DataSeries] where [end_date] is null)  " +
                    " ,@t  " +
                    " ,@p1 " +
                    " ,@p2";

                //paremeters are getting passe through the query

                cmd.Parameters.Add(new SqlParameter("@t", speed.timestamp));
                cmd.Parameters.Add(new SqlParameter("@p1", speed.speedValueMPH));
                cmd.Parameters.Add(new SqlParameter("@p2", speed.speedValueKPH));

                cmd.Connection = connection;

                try
                {
                    connection.Open();

                    if (cmd.ExecuteNonQuery() != 1)
                        SR3600Error("Error writing to the database.", null);
                    else
                        dataPoints++;

                    speedServer.NotifySubscribers(speed);
                }
                catch (Exception e)
                {
                    SR3600Error(e.Message, e);
                }
                finally
                {
                    connection.Close();
                }
            }
        }

        private void SR3600Error(string message, Exception e)
        {
            WriteToApplicationLog(message, EventLogEntryType.Information);

            if (e != null)
                WriteToApplicationLog(e.Message, EventLogEntryType.Error);

            if (e.InnerException != null)
                WriteToApplicationLog(e.InnerException.Message, EventLogEntryType.Error);

            if (errorCount++ > ERRORLIMIT)
            {
                WriteToApplicationLog("Too many errors. Shutting down service.", EventLogEntryType.Error);
                this.Stop();
            }
        }


        public static void WriteToApplicationLog(string message, EventLogEntryType type)
        {
			string sSource;
			string sLog;

			sSource = "SR3600 Data Collection Service";
			sLog = "Application";

			if (!EventLog.SourceExists(sSource))
				EventLog.CreateEventSource(sSource,sLog);

			EventLog.WriteEntry(sSource, 
                message, type);
		}
    }
}
