﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO.Ports;
using System.Globalization;
using System.Threading;

namespace DrHeet.SR3600Manager
{
    /// <summary>A speed data record as deliveded by the SR3600</summary>
    public delegate void SpeedDataHandler(SpeedRecord speed);

    /// <summary>The error handler delegate, to be called when an error is encountered.</summary>
    public delegate void ErrorHandler(string message, Exception e);

    /// <summary>The SpeedRecord struct holds a single measurement, along with timestamp and speed unit.</summary>
    public struct SpeedRecord
    {
        /// <summary>The timestamp is taken aftet the data was captured and right before the SpeedDataHandler delegate is called.</summary>
        public DateTime timestamp;
        /// <summary>The speed</summary>
        public double speedValueMPH;
        public double speedValueKPH;
        /// <summary>Speed can be recorded in MPH or KMH. The SR3600 uses MPH by default. This is only configurable on the device.</summary>
        public string speedUnit;
    }

    /// <summary>The SR3600Manager class encapsulates the serial communcation with the SR3600 device.</summary>
    public class SR3600Manager
    {
        #region PrivateParts

        // properties
        private string portName = "COM1";
        private int baudRate = 1200;
        private int dataBits = 8;
        private StopBits stopBits = StopBits.One;
        private Parity parity = Parity.None;
        private string speedUnit = "MPH";
        private char[] delimiterChars = {' '};
        private static int COMMANDWAIT = 100;

        // internal state
        private SerialPort comPort;             // keeps a reference to the com port
        private SpeedDataHandler callbackHandler;
        private ErrorHandler errorHandler;

        private CultureInfo culture;            // for SR3600 number format parsing

        private SpeedRecord sr3600Data;

        private void InitComPort()
        {
            comPort.PortName = portName;
            comPort.BaudRate = baudRate;
            comPort.DataBits = dataBits;
            comPort.StopBits = stopBits;
            comPort.Parity = parity;

            // the rest is not configurable for now
            comPort.Handshake = System.IO.Ports.Handshake.None;
            comPort.DtrEnable = true;
            comPort.RtsEnable = true;
            comPort.ReadTimeout = 1000;
            comPort.DataReceived += PortDataReceived;
            comPort.ReceivedBytesThreshold = 14;            // make sure there is always enough data to read
        }

        #endregion

        #region PublicParts

        /// <summary>Locale for number formatting</summary>
        public CultureInfo Culture
        {
            get { return culture; }
            set { culture = value; }
        }

        /// <summary>Name of the port the VK011 module is connected to (e.g. COM1, COM2, COM3, etc).</summary>
        public string PortName
        {
            get { return portName; }
            set { portName = value; }
        }

        /// <summary>The VK011 requires 9600 baud. Other speeds will not work.</summary>
        public int BaudRate
        {
            get { return baudRate; }
            set { baudRate = value; }
        }

        /// <summary>The VK011 requires 8 data bits.</summary>
        public int DataBits
        {
            get { return dataBits; }
            set { dataBits = value; }
        }

        /// <summary>The VK011 requires 1 stop bit (StopBits.One).</summary>
        public StopBits StopBits
        {
            get { return stopBits; }
            set { stopBits = value; }
        }

        /// <summary>The VK011 requires no parity (Parity.None).</summary>
        public Parity Parity
        {
            get { return parity; }
            set { parity = value; }
        }

        /// <summary>DegC or DeF - the VK011Manager by default records temperatures in Celsius (the SI unit for temperature).</summary>
        public string SpeedUnit
        {
            get { return speedUnit; }
            set
            {
                if (value.Equals("MPH") || value.Equals("KPH"))
                    speedUnit = value;
            }
        }

        /// <summary>Specifies a TemperatureDataHandler delegate to be called when a complete data set is ready for processing.</summary>
        public SpeedDataHandler CallBackHandler
        {
            set { callbackHandler = value; }
        }

        /// <summary>Specifies an ErrorHandler delegate to be called when a serial communication error occurs.</summary>
        public ErrorHandler OnError
        {
            set { errorHandler = value; }
        }

        /// <summary>This constructor takes a SerialPort object that can be instantiated externally.</summary>
        public SR3600Manager(SerialPort newPort)
        {
            comPort = newPort;
        }

        /// <summary>This constructor creates its own SerialPort object with default values.</summary>
        public SR3600Manager()
        {
            comPort = new SerialPort();
        }

        public bool InitSR3600()
        {
            bool success = true;

            try
            {
                InitComPort();
                Thread.Sleep(COMMANDWAIT);
                comPort.Open();
            }
            catch (Exception e)
            {
                if (errorHandler != null)
                    errorHandler(string.Format("Error initializing the serial port {0}: {1}\n", portName, e.Message), e);
            }

            return success;
        }


        /// <summary>This method closes the SerialPort.</summary>
        public void CleanupSR3600()
        {
            comPort.Close();
        }

        /// <summary>
        /// This method reads data from the SR3600 and parses it into a speedRecord structure.
        /// Data format:
        ///     x Sports Radar use only, Hex value          \  Word 0
        ///     0x7E ASCII Tilda                            /
        ///     0x20 ASCII Space
        ///     H Hundreds digit in ASCII (0x30 to 0x32)    \  Word 1
        ///     T Tens digit in ASCII (0x30 to 0x39)        |
        ///     O Ones digit in ASCII (0x30 to 0x39)        |
        ///     0x2E ASCII Period                           /
        ///     0x20 ASCII Space
        ///     0x4D or 0x4B ASCII M or K depending on display units selected   \  Word 2
        ///     0x50 ASCII P                                                    |
        ///     0x48 ASCII H                                                    |
        ///     x Sports Radar use only, Hex value                              |  There is likely another space here
        ///     0x0A ASCII Line feed                                            |
        ///     0x0D ASCII Carriage return                                      /
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void PortDataReceived(object sender, SerialDataReceivedEventArgs e)
        {

            try
            {
                string line = comPort.ReadLine();

                while (line.Equals(""))        // skip empty lines
                    line = comPort.ReadLine();

                string[] words = line.Split(delimiterChars);

                if (words.Length >= 3)
                {
                    sr3600Data.speedUnit = words[2].Substring(0, 3);
                    sr3600Data.timestamp = DateTime.Now;

                    if (sr3600Data.speedUnit.Equals("MPH"))
                    {
                        sr3600Data.speedValueMPH = float.Parse(words[1].Replace('?', ' '), culture);
                        sr3600Data.speedValueKPH = sr3600Data.speedValueMPH * 1.61;
                    }
                    else
                    {
                        sr3600Data.speedValueKPH = float.Parse(words[1].Replace('?', ' '), culture);
                        sr3600Data.speedValueMPH = sr3600Data.speedValueKPH / 1.61;
                    }


                    if (callbackHandler != null)
                        callbackHandler(sr3600Data);

                    sr3600Data.speedValueMPH = .0;
                    sr3600Data.speedValueKPH = .0;
                }
                else
                    if (errorHandler != null)
                        errorHandler("Unable to parse SR3600 data: [" + line + "]", new Exception("SR3600Manager: Incomplete or invalid data record detected."));
            }
            catch (Exception ex)
            {
                if (errorHandler != null)
                    errorHandler("Error retrieving or parsing SR3600 data", ex);
            }

        }

        #endregion
    }
}
